import 'dart:async';
import 'dart:convert';

import 'package:flutter/widgets.dart';
import 'package:logging/logging.dart';
import 'package:sunny_essentials/sunny_essentials.dart';
import 'package:sunny_sdk_core/model_exports.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';
import 'package:sunny_service_stubs/meta.dart';
import 'package:sunny_service_stubs/models/m_entity/m_entity.dart';

class MSchemaService with LifecycleAwareMixin implements IMSchemaService {
  static final Logger _log = Logger("mschemaService");
  Logger get log => _log;
  final ApiClient apiClient;
  final Map<MSchemaRef?, IMSchemaDefinition> _schemas = {};
  final Map<MOperationRef?, IMOperationDefinition> _operations = {};

  final StreamController<IMSchemaDefinition> _loadedSchemaCtrl =
      StreamController.broadcast();

  final SafeCompleter _load = SafeCompleter();

  MSchemaService(this.apiClient) {
    _loadAll();

    onDestroy('close stream', () => _loadedSchemaCtrl.close());
  }

  Stream<IMSchemaDefinition> get loadedSchemas async* {
    for (var schema in _schemas.values) {
      yield schema;
    }

    await for (var schema in _loadedSchemaCtrl.stream) {
      yield schema;
    }
  }

  Stream<IMEntityDefinition> get loadedEntitySchemas {
    return loadedSchemas.whereType();
  }

  void register(MSchemaRef schemaRef, IMEntityDefinition definition) {
    assert(schemaRef == definition.schemaRef,
        "$schemaRef == ${definition.schemaRef}");
    _internalSet(schemaRef, definition);
  }

  @override
  Future<IMSchemaDefinition> getSchema(MSchemaRef? ref,
      {bool reload = false}) async {
    if (reload != true && _schemas.containsKey(ref)) {
      return Future.value(_schemas[ref]);
    }
    return await _cacheSchema(ref!);
  }

  FutureOr<IMSchemaDefinition>? _internalGet(MSchemaRef ref,
      {bool reload = false}) {
    if (reload != true && _schemas.containsKey(ref)) {
      return _schemas[ref];
    }
    return _cacheSchema(ref);
  }

  FutureOr<IMEntityDefinition> getEntityDefinition(MSchemaRef ref,
      {bool reload = false}) {
    final schema = _internalGet(ref)!;
    return schema.thenOr((resolved) {
      if (schema is! IMEntityDefinition)
        throw "Schema $ref was not an entity schema";
      return schema;
    });
  }

  Future<List<ValidationError>> validate(MModel model) async {
    final ref = model.mtype;
    return await _validate(model, ref);
  }

  Future<IMOperationDefinition> getOperation(MOperationRef ref,
      {bool reload = false}) async {
    //TODO:ericm deal with
//    if (ref == userTaskOperationRef) return null;
    // ignore: unnecessary_null_comparison

    if (reload != true && _operations.containsKey(ref)) {
      final operation = _operations[ref]!;
      await _resolveOpSchemas(operation);
      return operation;
    } else {
      final cached = await _cacheOperation(ref);
      await _resolveOpSchemas(cached);
      return cached;
    }
  }

  @override
  IMSchemaDefinition tryGetSchema(MSchemaRef ref) {
    final existing = _schemas[ref] ?? (throw Exception("Missing schema for ref $ref"));
    return existing;
  }

  IMOperationDefinition? tryGetOpSchema(MOperationRef ref) {
    final existing = _operations[ref];
    return existing;
  }

  Future<List<IMOperationDefinition>> get assistedOperations async {
    await _load.future;
    return _operations.values.where((op) => op.isAssisted!).toList();
  }

  // ignore: unused_element
  Future _loadAll() async {
    for (var i = 0; i < 5; i++) {
      try {
        await _loadAllSchemas();
        await _loadAllOperations();
        _load.complete();
        return;
      } catch (e, stack) {
        if (i >= 4) {
          log.shout("Failed to load schemas: $e", e, stack);
          _load.complete();
          return;
        } else {
          log.warning("Failed to load schema $i - $e");
        }
      }
      await (2 * i).seconds.pause();
    }
  }

  Future _loadAllSchemas() async {
    final value = await apiClient.makeRequest("/schemas");

    final schemas = IMSchemaDefinitionList.fromJson(value);
    for (final schema in schemas.data.orEmpty()) {
      if (schema is IMEntityDefinition) {
        _internalSet(schema.schemaRef, schema);
      }
    }
  }

  void _internalSet(MSchemaRef? ref, IMSchemaDefinition definition) {
    _schemas[ref] = definition;
    _loadedSchemaCtrl.add(definition);
  }

  Future _loadAllOperations() async {
    var ops = await apiClient.makeRequest("/operations");

    // final value = json.decode('{}');
    final schemas = IMOperationDefinitionList.fromJson(ops);
    for (final operation in schemas.data.orEmpty()) {
      try {
        _operations[operation.operationRef] = operation;
        await _resolveOpSchemas(operation);
      } on ApiResponseException catch (e) {
        if (e.payload.errorType != "notFound") {
          log.severe(
              "Unable to load operation ${operation.operationRef}: $e", e);
        } else {
          log.fine("Not registering ${operation.operationRef}  $e");
        }
      } catch (e, stack) {
        log.severe("Error registering all ops ${e}", e, stack);
      }
    }
  }

  Future<IMSchemaDefinition> _cacheSchema(MSchemaRef ref,
      {bool resolveProps = true}) async {
    log.fine("Fetching mschema for $ref");

    final value = await apiClient.makeRequest(ref.relativeSchemaUri,
        onError: (code, body) {
      if (code != 404) return null;
      log.info("Missing schema in ${ref.baseCode}");
      return emptySchema(ref);
    });

    log.fine("Caching mschema for $ref");
    final definition = IMSchemaDefinition.fromJson(value);
    _schemas[ref] = definition;
    if (resolveProps) {
      _resolveProperties(definition);
    }

    _internalSet(ref, definition);
    return definition;
  }

  Future _resolveProperties(IMSchemaDefinition definition) async {
    for (final prop in definition.properties!) {
      if (prop is IEmbeddedProperty) await getSchema(prop.embedRef);
    }
  }

  Future<IMOperationDefinition> _cacheOperation(MOperationRef ref) async {
    log.info("Fetching operation for $ref");

    apiClient.makeRequest(ref.relativeSchemaUri, onError: (status, _) {});
//    var client = apiClient;
//    final response = await client.invokeAPI(RequestBuilder()
//      ..path =
//      ..server = ServerInfo(client.config.mEntityBaseUrl)
//      ..contentType = "text/plain"
//      ..method = HttpMethod.GET);
//    if (response.statusCode != 200) {
//      throw ApiException.response(response.statusCode, response.body);
//    }
//    final value = json.decode(response.body);
    final value = json.decode('{}');
    log.fine("Caching mschema for $ref");
    final definition = IMOperationDefinition.fromJson(value);
    _operations[ref] = definition;

    _resolveOpSchemas(definition);
    return definition;
  }

  Future<List<ValidationError>> _validate(MModel model, MSchemaRef ref) async {
    log.info("Validating model for $ref");

    var client = apiClient;
    final response = await client.invokeRequest(RequestBuilder()
      ..path = "${ref.relativeSchemaBaseUri}/validate"
      ..basePath = apiClient.basePath
      ..contentType = "application/json"
      ..body = model.wrapped
      ..method = HttpMethod.POST);
    if (response.statusCode != 200) {
      throw ApiException.response(response.statusCode, response.body);
    }
    final value = (json.decode(response.body) as List?) ?? [];
    return [...value.map((json) => ValidationError.ofJson(json))];
  }

  Future _resolveOpSchemas(IMOperationDefinition definition) async {
    if (definition.initSchema is IMRefSchemaDefinition) {
      definition.initSchema = await _resolve(definition.initSchema);
    }
    if (definition.inputSchema is IMRefSchemaDefinition) {
      definition.inputSchema = await _resolve(definition.inputSchema);
    }
    if (definition.outputSchema is IMRefSchemaDefinition) {
      definition.outputSchema = await _resolve(definition.outputSchema);
    }
  }

  Future<IMSchemaDefinition?> _resolve(IMSchemaDefinition? definition) async {
    if (definition is! IMRefSchemaDefinition) return definition;
    final ref = definition;
    return await getSchema(ref.schemaRef);
  }

  bool get isLoaded => _load.isCompleted;

  Future load() => _load.future;
}

IMSchemaDefinition emptySchema(MSchemaRef ref) {
  return IMSchemaDefinition.of(
    schemaURI: ref.fullUri.toUri(),
    properties: [],
  );
}

extension BuildContextMSchemaService on BuildContext {
  Future<IMSchemaDefinition> loadSchema(MSchemaRef ref) async {
    return this.consume<IMSchemaService>()!.getSchema(ref);
  }

  MSchemaService get mschemaService => sunny.get<MSchemaService>();
}

extension MSchemaLoadersExt on MModel {
  /// Fetches (or uses a cached) schema from the server.  This would be a good target for an extension function
  /// if/when dart does that
  Future<IMSchemaDefinition> getMSchema(BuildContext context) async {
    return await context.consume<IMSchemaService>()!.getSchema(this.mtype);
  }

  IMSchemaDefinition getMSchemaOrNull(BuildContext context) =>
      sunny.get<MSchemaService>().tryGetSchema(this.mtype);

  Future<IMSchemaDefinition> refreshSchema(BuildContext context) =>
      context.consume<IMSchemaService>()!.getSchema(this.mtype, reload: true);
}

extension ApiClientFetchExt on ApiClient {
  Future<dynamic> loadSchema(MSchemaRef ref) {
    return makeRequest(ref.relativeSchemaUri);
  }

  Future<dynamic> makeRequest(url,
      {dynamic onError(status, String body)?}) async {
    final response = await invokeRequest(RequestBuilder()
      ..path = "$url"
      ..basePath = basePaths["mEntityBaseUrl"]
      ..contentType = "application/json"
      ..method = HttpMethod.GET);

    if (response.statusCode != 200) {
      final resp = onError?.call(response.statusCode, response.body);
      if (resp == null) {
        throw ApiException.response(response.statusCode, response.body);
      }
      return resp;
    }
    return json.decode(response.body);
  }
}

extension MSchemaLoaders2 on MSchemaRef {
  /// Fetches (or uses a cached) schema from the server.  This would be a good target for an extension function
  /// if/when dart does that
  Future<IMSchemaDefinition> getMSchema() async {
    return await sunny.get<MSchemaService>().getSchema(this);
  }

  IMSchemaDefinition getMSchemaOrNull(BuildContext context) =>
      sunny.get<MSchemaService>().tryGetSchema(this);

  Future<IMSchemaDefinition> refreshSchema(BuildContext context) =>
      sunny.get<MSchemaService>().getSchema(this, reload: true);
}
